import unittest
from unittest.mock import patch, mock_open, MagicMock
import urllib.error
import warnings
import json
from smufolib.request import Request, writeJson, CONFIG
from smufolib.error import URLWarning


# pylint: disable=W0212, W0613


class TestRequest(unittest.TestCase):
    def setUp(self):
        self.path = "http://example.com"
        self.fallback = "fallback.txt"
        self.request = Request(path=self.path, fallback=self.fallback, warn=False)

    @patch("urllib.request.urlopen")
    def test_readFromURL(self, mock_urlopen):
        mock_response = MagicMock()
        mock_response.read.return_value = b"data from URL"
        mock_urlopen.return_value.__enter__.return_value = mock_response

        result = self.request._readFromURL()
        self.assertEqual(result, b"data from URL")

    @patch("builtins.open", new_callable=mock_open, read_data="data from file")
    def test_readFromFallback(self, mock_file):
        result = self.request._readFromFallback()
        self.assertEqual(result, "data from file")

    @patch("builtins.open", new_callable=mock_open, read_data="data from file")
    def test_readFromPath(self, mock_file):
        self.request._path = "path.txt"
        result = self.request._readFromPath()
        self.assertEqual(result, "data from file")

    @patch("urllib.request.urlopen", side_effect=urllib.error.URLError("URL error"))
    @patch("builtins.open", new_callable=mock_open, read_data="data from fallback")
    def test_handleURLError_with_fallback(self, mock_file, mock_urlopen):
        result = self.request._readFromURL()
        self.assertEqual(result, "data from fallback")

    @patch("urllib.request.urlopen", side_effect=urllib.error.URLError("URL error"))
    def test_handleURLError_without_fallback(self, mock_urlopen):
        self.request._fallback = None
        with self.assertRaises(urllib.error.URLError):
            self.request._readFromURL()

    @patch("urllib.request.urlopen", side_effect=urllib.error.URLError("URL error"))
    @patch("builtins.open", new_callable=mock_open, read_data="data from fallback")
    def test_handleURLError_with_warning(self, mock_file, mock_urlopen):
        self.request._warn = True
        with warnings.catch_warnings(record=True) as w:
            result = self.request._readFromURL()
            self.assertEqual(result, "data from fallback")
            self.assertTrue(any(item.category == URLWarning for item in w))

    @patch("builtins.open", new_callable=mock_open, read_data="data from file")
    def test_raw(self, mock_file):
        self.request._path = "path.txt"
        result = self.request.raw
        self.assertEqual(result, "data from file")

    @patch("builtins.open", new_callable=mock_open, read_data='{"key": "value"}')
    def test_json(self, mock_file):
        self.request._path = "path.json"
        result = self.request.json()
        self.assertEqual(result, {"key": "value"})

    def test_path(self):
        self.assertEqual(self.request.path, self.path)

    def test_fallback(self):
        self.assertEqual(self.request.fallback, self.fallback)

    def test_mode(self):
        self.assertEqual(self.request.mode, "r")

    def test_encoding(self):
        self.assertEqual(self.request.encoding, CONFIG["request"]["encoding"])


class TestWriteJson(unittest.TestCase):
    @patch("builtins.open", new_callable=mock_open)
    def test_writeJson(self, mock_file):
        filepath = "test.json"
        source = {"key": "value"}
        writeJson(filepath, source)
        mock_file.assert_called_once_with(
            filepath, "w", encoding=CONFIG["request"]["encoding"]
        )
        handle = mock_file()
        written_content = "".join(call.args[0] for call in handle.write.call_args_list)
        expected_content = json.dumps(source, indent=4, sort_keys=False)
        self.assertEqual(written_content, expected_content)

    def test_writeJson_invalid_type(self):
        with self.assertRaises(TypeError):
            writeJson(123, {"key": "value"})

    def test_writeJson_invalid_extension(self):
        with self.assertRaises(ValueError):
            writeJson("test.txt", {"key": "value"})

    @patch("builtins.open", new_callable=mock_open)
    def test_writeJson_serialization_error(self, mock_file):
        source = {"key": set()}
        with self.assertRaises(ValueError):
            writeJson("test.json", source)



import unittest
from smufolib.stdUtils import flatten, addTuples, isFloat, validateClassAttr

# pylint: disable=C0115, C0116, C0103


class StdUtils(unittest.TestCase):
    def test_flatten(self):
        matrix = [[[i for i in range(5)] for j in range(5)] for h in range(5)]
        self.assertEqual(list(flatten(matrix)), list(range(5)) * 25)
        self.assertEqual(
            list(flatten(matrix, 1)), [[i for i in range(5)] for j in range(25)]
        )

    def test_addTuples(self):
        self.assertEqual(addTuples((2, 4), (2, 4)), (4, 8))

    def test_isFloat(self):
        self.assertTrue(isFloat("1.2"))
        self.assertFalse(isFloat("1"))

    def test_validateClassAttr(self):
        with self.assertRaises(TypeError):
            validateClassAttr(object, 42)



import sys
import unittest
from tests import test_cli
from tests import test_config
from tests import test_converters
from tests import test_engravingDefaults
from tests import test_error
from tests import test_normalizers
from tests import test_pointUtils
from tests import test_range
from tests import test_request
from tests import test_smufl
from tests import test_stdUtils


def testEnvironment(objectGenerator, inApp=False, verbosity=1, testNormalizers=True):
    modules = [
        test_cli,
        test_config,
        test_converters,
        test_engravingDefaults,
        test_error,
        test_pointUtils,
        test_range,
        test_request,
        test_smufl,
        test_stdUtils,
    ]
    if testNormalizers:
        modules.append(test_normalizers)

    globalSuite = unittest.TestSuite()
    loader = unittest.TestLoader()
    for module in modules:
        suite = loader.loadTestsFromModule(module)
        _setObjectGenerator(suite, objectGenerator)
        globalSuite.addTest(suite)
    runner = unittest.TextTestRunner(verbosity=verbosity)
    succes = runner.run(globalSuite).wasSuccessful()
    if not inApp:
        sys.exit(not succes)
    else:
        return succes  # pragma: no cover


def _setObjectGenerator(suite, objectGenerator):
    for i in suite:
        if isinstance(i, unittest.TestSuite):
            _setObjectGenerator(i, objectGenerator)
        else:
            i.objectGenerator = objectGenerator



#!/usr/bin python3
# coding: utf-8

"""
This script calculates and sets attribute values for
the :class:`.EngravingDefaults` class based on glyph dimensions,
registration, metrics, and point placement.

The script provides options to override automatic contour measurements
and reassign attributes to different ruler functions or glyphs. It aims
to automate the process, but due to limitations in parameter
identification and control, it should only be used as a starting point.

By default, attributes are mapped to the following functions and
glyphs:

    ===================================   ====================   ===========
    Attribute                             Function               Glyph
    ===================================   ====================   ===========
    :attr:`.arrowShaftThickness`          :func:`xOrigin`        `uniEB60`
    :attr:`.barlineSeparation`            :func:`xInner`         `uniE031`
    :attr:`.beamSpacing`                  :func:`yInner`         `uniE1F9`
    :attr:`.beamThickness`                :func:`boundsHeight`   `uniE1F7`
    :attr:`.bracketThickness`             :func:`xOrigin`        `uniE003`
    :attr:`.dashedBarlineDashLength`      :func:`yMinimum`       `uniE036`
    :attr:`.dashedBarlineGapLength`       :func:`yInner`         `uniE036`
    :attr:`.dashedBarlineThickness`       :func:`xOrigin`        `uniE036`
    :attr:`.hairpinThickness`             :func:`yMinimum`       `uniE53E`
    :attr:`.hBarThickness`                :func:`yMinimum`       `uniE4F0`
    :attr:`.legerLineExtension`           :func:`boundsLeft`     `uniE022`
    :attr:`.legerLineThickness`           :func:`boundsHeight`   `uniE022`
    :attr:`.lyricLineThickness`           :func:`boundsHeight`   `uniE010`
    :attr:`.octaveLineThickness`          :func:`boundsHeight`   `uniE010`
    :attr:`.pedalLineThickness`           :func:`boundsHeight`   `uniE010`
    :attr:`.repeatBarlineDotSeparation`   :func:`stemDot`        `uniE040`
    :attr:`.repeatEndingLineThickness`    :func:`xOrigin`        `uniE030`
    :attr:`.slurEndpointThickness`        :func:`xOrigin`        `uniE1FD`
    :attr:`.slurMidpointThickness`        :func:`yMinimum`       `uniE1FD`
    :attr:`.staffLineThickness`           :func:`boundsHeight`   `uniE010`
    :attr:`.stemThickness`                :func:`xOrigin`        `uniE210`
    :attr:`.subBracketThickness`          :func:`xOrigin`        `uniE030`
    :attr:`.textEnclosureThickness`       :func:`boundsHeight`   `uniE010`
    :attr:`.thickBarlineThickness`        :func:`xOrigin`        `uniE034`
    :attr:`.thinBarlineThickness`         :func:`xOrigin`        `uniE030`
    :attr:`.thinThickBarlineSeparation`   :func:`xInner`         `uniE032`
    :attr:`.tieEndpointThickness`         :func:`xOrigin`        `uniE1FD`
    :attr:`.tieMidpointThickness`         :func:`yMinimum`       `uniE1FD`
    :attr:`.tupletBracketThickness`       :func:`xOrigin`        `uniE1FE`
    ===================================   ====================   ===========

.. note:: The setting for :attr:`.textFontFamily` must be set manually
   within the `override` parameter.

This script requires SMufoLib to be installed within its executive
environment. It may also be imported as a module and contains the
following public functions:

    - :func:`calculateEngravingDefaults` - The scripts program function.
    - :func:`main` – Command line entry point.
    - :func:`boundsLeft` – Returns absolute value of bounds x minimum.
    - :func:`boundsHeight` – Returns absolute value of bounds height.
    - :func:`stemDot` – Measures distance between stem and dot
      countour.
    - :func:`xInner` – Measures distance between two adjacent x points
      of different contours.
    - :func:`xOrigin` – Measures distance between two adjacent x points
      closest to origin.
    - :func:`yInner` – Measures distance between two adjacent y points
      of different contours.
    - :func:`yMinimum` – Measures distance between two adjacent
      low-points on y axis.

For command-line options, run the script with :option:`--help`
argument.

"""

from __future__ import annotations
from typing import TYPE_CHECKING
from collections.abc import Callable
import argparse
import json
import textwrap
from pathlib import Path

from tqdm import tqdm

from smufolib import (
    Font,
    cli,
    config,
    converters,
    error,
    normalizers,
    pointUtils,
    stdUtils,
)

if TYPE_CHECKING:
    from fontParts.fontshell.point import RPoint
    from smufolib.objects.glyph import Glyph

Exclude = tuple[str] | list[str]
OverrideValue = int | float | tuple[str, ...] | None
Override = dict[str, OverrideValue]
MappingValue = str | int
Mapping = dict[str, MappingValue]
Remapping = dict[str, Mapping]

CONFIG = config.load()

# Parameter defaults
EXCLUDE = None
OVERRIDE = None
SPACES = False
REMAP = None
MARGIN_OF_ERROR = 6
VERBOSE = False
MAPPING = {
    "arrowShaftThickness": {
        "ruler": "xOrigin",
        "glyph": "uniEB60",
        "referenceIndex": 0,
    },
    "barlineSeparation": {"ruler": "xInner", "glyph": "uniE031", "referenceIndex": 3},
    "beamSpacing": {"ruler": "yInner", "glyph": "uniE1F9", "referenceIndex": 3},
    "beamThickness": {"ruler": "boundsHeight", "glyph": "uniE1F7", "referenceIndex": 0},
    "bracketThickness": {"ruler": "xOrigin", "glyph": "uniE003", "referenceIndex": 0},
    "dashedBarlineDashLength": {
        "ruler": "yMinimum",
        "glyph": "uniE036",
        "referenceIndex": 0,
    },
    "dashedBarlineGapLength": {
        "ruler": "yInner",
        "glyph": "uniE036",
        "referenceIndex": 3,
    },
    "dashedBarlineThickness": {
        "ruler": "xOrigin",
        "glyph": "uniE036",
        "referenceIndex": 0,
    },
    "hairpinThickness": {"ruler": "yMinimum", "glyph": "uniE53E", "referenceIndex": 0},
    "hBarThickness": {"ruler": "yMinimum", "glyph": "uniE4F0", "referenceIndex": 0},
    "legerLineExtension": {
        "ruler": "boundsLeft",
        "glyph": "uniE022",
        "referenceIndex": 0,
    },
    "legerLineThickness": {
        "ruler": "boundsHeight",
        "glyph": "uniE022",
        "referenceIndex": 0,
    },
    "lyricLineThickness": {
        "ruler": "boundsHeight",
        "glyph": "uniE010",
        "referenceIndex": 0,
    },
    "octaveLineThickness": {
        "ruler": "boundsHeight",
        "glyph": "uniE010",
        "referenceIndex": 0,
    },
    "pedalLineThickness": {
        "ruler": "boundsHeight",
        "glyph": "uniE010",
        "referenceIndex": 0,
    },
    "repeatBarlineDotSeparation": {
        "ruler": "stemDot",
        "glyph": "uniE040",
        "referenceIndex": 0,
    },
    "repeatEndingLineThickness": {
        "ruler": "xOrigin",
        "glyph": "uniE030",
        "referenceIndex": 0,
    },
    "slurEndpointThickness": {
        "ruler": "xOrigin",
        "glyph": "uniE1FD",
        "referenceIndex": 0,
    },
    "slurMidpointThickness": {
        "ruler": "yMinimum",
        "glyph": "uniE1FD",
        "referenceIndex": 0,
    },
    "staffLineThickness": {
        "ruler": "boundsHeight",
        "glyph": "uniE010",
        "referenceIndex": 0,
    },
    "stemThickness": {"ruler": "xOrigin", "glyph": "uniE210", "referenceIndex": 0},
    "subBracketThickness": {
        "ruler": "xOrigin",
        "glyph": "uniE030",
        "referenceIndex": 0,
    },
    "textEnclosureThickness": {
        "ruler": "boundsHeight",
        "glyph": "uniE010",
        "referenceIndex": 0,
    },
    "textFontFamily": {},
    "thickBarlineThickness": {
        "ruler": "xOrigin",
        "glyph": "uniE034",
        "referenceIndex": 0,
    },
    "thinBarlineThickness": {
        "ruler": "xOrigin",
        "glyph": "uniE030",
        "referenceIndex": 0,
    },
    "thinThickBarlineSeparation": {
        "ruler": "xInner",
        "glyph": "uniE032",
        "referenceIndex": 3,
    },
    "tieEndpointThickness": {
        "ruler": "xOrigin",
        "glyph": "uniE1FD",
        "referenceIndex": 0,
    },
    "tieMidpointThickness": {
        "ruler": "yMinimum",
        "glyph": "uniE1FD",
        "referenceIndex": 0,
    },
    "tupletBracketThickness": {
        "ruler": "xOrigin",
        "glyph": "uniE1FE",
        "referenceIndex": 0,
    },
}


# pylint: disable=R0913, C0103, R0914


def calculateEngravingDefaults(
    font: Font | Path | str,
    exclude: Exclude | None = EXCLUDE,
    override: Override | None = OVERRIDE,
    remap: Remapping | None = REMAP,
    spaces: bool = SPACES,
    verbose: bool = VERBOSE,
) -> None:
    """Calculate engraving defaults from glyph contours.

    :param font: Target font object or path to file.
    :param exclude: :class:`.EngravingDefaults` attributes to exclude.
        Defaults to :obj:`None`.
    :param override: :class:`.EngravingDefaults` attributes to manually
        override mapped to their values. Defaults to :obj:`None`.
    :param remap: :class:`.EngravingDefaults` attributes mapped to
        remappings :class:`dict`, e.g.:

        .. code-block:: python

            r = {
                'arrowShaftThickness': {
                    'ruler': 'boundsLeft',
                    'glyph': 'uniEB60',
                    'referenceIndex': 2
                }
            }

        Defaults to :obj:`None`.
    :param spaces: Whether values for overrides are given in staff
        spaces as opposed to font units. Defaults to :obj:`False`.
    :param verbose: Make output verbose. Defaults to :obj:`False`.
    :raises TypeError: If any parameter value is not the expected type.
    :raises ValueError: If any parameter value item is not the expected
        type or value.

    """

    dispatcher = {
        "boundsHeight": boundsHeight,
        "boundsLeft": boundsLeft,
        "stemDot": stemDot,
        "xInner": xInner,
        "xOrigin": xOrigin,
        "yInner": yInner,
        "yMinimum": yMinimum,
    }

    print("Starting...")

    font = _normalizeFont(font)
    exclude = _normalizeExclude(exclude)
    override = _normalizeOverride(override)
    remap = _normalizeRemap(remap)

    font.smufl.spaces = False

    iterator = MAPPING.items()
    if not verbose:
        iterator = tqdm(MAPPING.items())

    stdUtils.verbosePrint("\nSetting attributes:", verbose)
    for key, mapping in iterator:
        if exclude:
            if key in exclude:
                continue

        if key == "textFontFamily":
            if override and key in override:
                value = override[key]
                setattr(font.smufl.engravingDefaults, key, value)
                stdUtils.verbosePrint(f"\t'{key}': {value}", verbose)
            continue

        ruler = mapping["ruler"]
        glyph = mapping["glyph"]
        referenceIndex = mapping["referenceIndex"]
        remapping = remap.get(key, {}) if remap else {}
        ruler = dispatcher[remapping.get("ruler", ruler)]  # type: ignore
        glyph = remapping.get("glyph", glyph)
        referenceIndex = remapping.get("referenceIndex", referenceIndex)

        rulerValue = _getValue(
            key=key,
            glyph=font[glyph],
            ruler=ruler,  # type: ignore
            referenceIndex=referenceIndex,  # type: ignore
            verbose=verbose,
        )
        if rulerValue is None:
            continue

        if override and key in override:
            value = override[key]
        else:
            value = rulerValue
        if spaces and isinstance(value, (int, float)):
            value = font.smufl.toUnits(value)

        setattr(font.smufl.engravingDefaults, key, value)
        if spaces and value and isinstance(value, (int, float)):
            value = font.smufl.toSpaces(value)
        stdUtils.verbosePrint(f"\t'{key}': {value}", verbose)

    stdUtils.verbosePrint("\nSaving font...", verbose)
    font.save()
    print("\nDone!")


def main() -> None:
    """Command line entry point."""
    args = _parseArgs()
    calculateEngravingDefaults(
        font=args.font,
        exclude=args.exclude,
        override=args.override,
        remap=args.remap,
        spaces=args.spaces,
        verbose=args.verbose,
    )


# ------
# Rulers
# ------


def boundsHeight(glyph: Glyph) -> int | float | None:
    """Height of the glyph bounds.

    :param glyph: Source :class:`.Glyph` of contours to measure.

    """
    return converters.toIntIfWhole(glyph.bounds[3] - glyph.bounds[1])


def boundsLeft(glyph: Glyph) -> int | float | None:
    """Return absolute value of the glyph bounds x minimum value.

    :param glyph: Source :class:`.Glyph` of contours to measure.

    """
    return converters.toIntIfWhole(abs(glyph.bounds[0]))


def stemDot(glyph: Glyph, referenceIndex: int = 0) -> int | float | None:
    """Calculate distance between stem and dot countour.

    :param glyph: Source :class:`.Glyph` of contours to measure.
    :param referenceIndex: referenceIndex of reference point. Defaults
        to ``0``.

    """
    curves = sorted(pointUtils.getPoints(glyph, "curve"), key=lambda p: p.position.x)
    lines = sorted(
        pointUtils.getPoints(glyph, "line"), reverse=True, key=lambda p: p.position.x
    )

    reference = curves[referenceIndex]

    for point in lines:
        if not point.contourIndex != reference.contourIndex:
            continue
        return converters.toIntIfWhole(abs(point.position.x - reference.position.x))

    return None


def xInner(glyph: Glyph, referenceIndex: int = 0) -> int | float | None:
    """Calculate distance between adjacent x points of two contours.

    :param glyph: Source :class:`.Glyph` of contours to measure.
    :param referenceIndex: referenceIndex of reference point. Defaults
        to ``0``.

    """
    points = sorted(pointUtils.getPoints(glyph), key=lambda p: p.position.x)
    reference = points[referenceIndex]

    for point in points:
        if not (
            point.contourIndex != reference.contourIndex
            and _areAdjacent(point, reference, axis="y")
        ):
            continue
        return converters.toIntIfWhole(abs(point.position.x - reference.position.x))

    return None


def xOrigin(glyph: Glyph, referenceIndex: int = 0) -> int | float | None:
    """Calculate distance between adjacent x points closest to origin.

    :param glyph: Source :class:`.Glyph` of contours to measure.
    :param referenceIndex: referenceIndex of reference point. Defaults
        to ``0``.

    """
    points = sorted(pointUtils.getPoints(glyph), key=lambda p: sum(p.position))
    reference = points[referenceIndex]

    for point in points:
        if not (
            point.position.x != reference.position.x
            and point.contourIndex == reference.contourIndex
            and _areAdjacent(point, reference, axis="y")
        ):
            continue
        return converters.toIntIfWhole(abs(point.position.x - reference.position.x))

    return None


def yOrigin(glyph: Glyph, referenceIndex: int = 0) -> int | float | None:
    """Calculate distance between adjacent x points closest to origin.

    :param glyph: Source :class:`.Glyph` of contours to measure.
    :param referenceIndex: referenceIndex of reference point. Defaults
        to ``0``.

    """
    points = sorted(pointUtils.getPoints(glyph), key=lambda p: sum(p.position))
    reference = points[referenceIndex]

    for point in points:
        if not (
            point.position.x != reference.position.x
            and point.contourIndex == reference.contourIndex
            and _areAdjacent(point, reference, axis="y")
        ):
            continue
        return converters.toIntIfWhole(abs(point.position.x - reference.position.x))

    return None


def yInner(glyph: Glyph, referenceIndex: int = 0) -> int | float | None:
    """Calculate distance between adjacent y points of two contours.

    :param glyph: Source :class:`.Glyph` of contours to measure.
    :param referenceIndex: referenceIndex of reference point. Defaults
        to ``0``.

    """
    points = sorted(pointUtils.getPoints(glyph), key=lambda p: p.position.y)
    reference = points[referenceIndex]
    for point in points:
        if not (
            point.contourIndex != reference.contourIndex
            and _areAdjacent(point, reference, axis="x")
        ):
            continue
        return converters.toIntIfWhole(abs(point.position.y - reference.position.y))

    return None


def yMinimum(glyph: Glyph, referenceIndex: int = 0) -> int | float | None:
    """Calculate distance between adjacent low-points on axis y.

    :param glyph: Source :class:`.Glyph` of contours to measure.
    :param referenceIndex: referenceIndex of reference point. Defaults
        to ``0``.

    """
    points = sorted(pointUtils.getPoints(glyph), key=lambda p: p.position.y)
    reference = points[referenceIndex]

    for point in points:
        if not (
            point.position.y != reference.position.y
            and point.contourIndex == reference.contourIndex
            and _areAdjacent(point, reference, axis="x")
        ):
            continue
        return converters.toIntIfWhole(abs(point.position.y - reference.position.y))

    return None


# -------
# Helpers
# -------


def _normalizeFont(font: Font | Path | str) -> Font:
    # Convert font path to object if necessary.
    error.validateType(font, (Font, Path, str), "font")
    if isinstance(font, Font):
        return font
    return Font(font)


def _normalizeExclude(exclude: Exclude | None) -> Exclude | None:
    # Noralize `exclude`.
    if exclude is None:
        return None

    error.validateType(exclude, (tuple, list), "exclude")

    for item in exclude:
        error.suggestValue(item, list(MAPPING.keys()), "exclude", items=True)
    return exclude


def _normalizeOverride(override: Override | None) -> Override | None:
    # Normalize override dictionary.
    if override is None:
        return None

    error.validateType(override, dict, "override")
    for key, value in override.items():
        attributes = list(MAPPING.keys())
        error.suggestValue(key, attributes, "override", items=True)
        normalizers.normalizeEngravingDefaultsAttr(key, value)

    return override


def _normalizeRemap(remap: Remapping | None) -> Remapping | None:
    # Normalize remap dictionary.
    if remap is None:
        return None

    error.validateType(remap, dict, "remap")

    for key, value in remap.items():
        error.suggestValue(key, list(MAPPING.keys()), "remap", items=True)

        for k, v in value.items():
            error.suggestValue(
                k,
                ("ruler", "glyph", "referenceIndex"),
                f"'remap'['{value}']",
                items=True,
            )

            error.validateType(v, (str, int), f"'remap'['{value}']", items=True)

    return remap


def _getValue(
    key: str, glyph: Glyph, ruler: Callable, referenceIndex: int, verbose: bool
) -> int | float | None:
    # Get value from ruler function and print error message.
    try:
        if referenceIndex:
            return ruler(glyph, referenceIndex)
        return ruler(glyph)
    except KeyError:
        stdUtils.verbosePrint(
            "Skipping attribute assigned to non-"
            f"existent glyph: '{key}' ('{glyph.name}')",
            verbose,
        )
        return None


def _parseArgs() -> argparse.Namespace:
    # Parse command line arguments and options.
    parser = argparse.ArgumentParser(
        description=stdUtils.getSummary(calculateEngravingDefaults.__doc__),
        parents=[
            cli.commonParser(
                "font", exclude=EXCLUDE, spaces=SPACES, verbose=VERBOSE, addHelp=False
            )
        ],
        formatter_class=cli.createHelpFormatter(
            ("ArgumentDefaultsHelpFormatter", "RawDescriptionHelpFormatter")
        ),
    )

    parser.add_argument(
        "-o",
        "--override",
        default=OVERRIDE,
        type=json.loads,
        help=textwrap.dedent(
            """JSON string of attributes and values to manually override;
            in the format '{"<attribute name>": <value>, ...}'"""
        ),
    )

    parser.add_argument(
        "-r",
        "--remap",
        default=REMAP,
        type=json.loads,
        help=textwrap.dedent(
            """JSON string of ruler, glyph and referenceIndex remappings
            in the format: {"<attribute name>":
            '{"ruler": "<function name>", "glyph": "<glyph name>",
            "referenceIndex": <integer>}, ...}'"""
        ),
    )
    return parser.parse_args()


def _areAdjacent(point1: RPoint, point2: RPoint, axis: str | None) -> bool:
    # Check if points are adjacent on axis.
    # Employ margin of error for point placement.

    def withinRange(num1: int, num2: int) -> bool:
        return num2 - MARGIN_OF_ERROR <= num1 <= num2 + MARGIN_OF_ERROR

    x1, x2 = point1.position.x, point2.position.x
    y1, y2 = point1.position.y, point2.position.y

    if axis == "x":
        checkAxis = withinRange(x1, x2)
    elif axis == "y":
        checkAxis = withinRange(y1, y2)
    else:
        checkAxis = withinRange(x1, x2) or withinRange(y1, y2)
    return point1 != point2 and checkAxis


if __name__ == "__main__":
    main()



import unittest
from smufolib.pointUtils import (
    Point,
    Position,
    getPoints,
    getContourPoints,
    getCompositePoints,
)


class TestPointUtils(unittest.TestCase):
    def setUp(self):
        self.font, _ = self.objectGenerator("font")  # pylint: disable=E1101
        self.layer = self.font.newLayer("testLayer")
        self.font.defaultLayer = self.layer
        self.glyph1 = self.font.newGlyph("testGlyph1")
        self.addContour(self.glyph1, (1, 2), (3, 4))
        self.glyph2 = self.font.newGlyph("testGlyph2")
        self.addContour(self.glyph2, (5, 6), (7, 8))
        self.component = self.glyph1.appendComponent("testGlyph2", offset=(1, 1))

        # Expected points for contour and composite tests
        self.expectedContourPoints1 = (
            Point("line", position=Position(1, 2), contourIndex=0),
            Point("line", position=Position(3, 4), contourIndex=0),
        )

        self.expectedContourPoints2 = (
            Point("line", position=Position(5, 6), contourIndex=0),
            Point("line", position=Position(7, 8), contourIndex=0),
        )

        self.expectedCompositePoints = (
            Point("line", position=Position(6, 7), contourIndex=0),  # (5+1, 6+1)
            Point("line", position=Position(8, 9), contourIndex=0),  # (7+1, 8+1)
        )

    def addContour(self, glyph, *points):
        pen = glyph.getPen()
        pen.moveTo(points[0])
        for point in points[1:]:
            pen.lineTo(point)
        pen.closePath()

    def assertPointListsEqual(self, actual, expected):
        self.assertCountEqual(actual, expected)
        for point in actual:
            self.assertIsInstance(point, Point)
            self.assertIsInstance(point.position, Position)

    def test_getContourPoints_with_glyph(self):
        points = tuple(getContourPoints(self.glyph1))
        self.assertPointListsEqual(points, self.expectedContourPoints1)

    def test_getCompositePoints_with_glyph(self):
        points = tuple(getCompositePoints(self.glyph1))
        self.assertPointListsEqual(points, self.expectedCompositePoints)

    def test_getPoints_with_glyph(self):
        points = tuple(getPoints(self.glyph1))
        expectedPoints = self.expectedContourPoints1 + self.expectedCompositePoints
        self.assertPointListsEqual(points, expectedPoints)

    def test_getContourPoints_with_font(self):
        points = tuple(getContourPoints(self.font))
        expectedPoints = self.expectedContourPoints1 + self.expectedContourPoints2
        self.assertPointListsEqual(points, expectedPoints)

    def test_getCompositePoints_with_font(self):
        points = tuple(getCompositePoints(self.font))
        self.assertPointListsEqual(points, self.expectedCompositePoints)

    def test_getPoints_with_font(self):
        points = tuple(getPoints(self.font))
        expectedPoints = (
            self.expectedContourPoints1
            + self.expectedContourPoints2
            + self.expectedCompositePoints
        )
        self.assertPointListsEqual(points, expectedPoints)



#!/usr/bin/env python3
# coding: utf-8
"""
This script sets glyph anchors based on data from a SMuFL font metadata
JSON file (SMuFL's reference font, Bravura, by default).

This script requires SMufoLib to be installed within its
executive environment. It may also be imported as a module and contains
the following public funcitons:

    - :func:`importAnchors` – The scripts program function.
    - :func:`main` - Command line entry point.

"""

from __future__ import annotations
from typing import Any
import argparse
from contextlib import nullcontext
from pathlib import Path
import time

from tqdm import tqdm

from smufolib.objects.font import Font
from smufolib.request import Request
from smufolib import cli, config, error, normalizers, stdUtils

JsonDict = dict[str, Any]
ColorValue = int | float
ColorTuple = tuple[ColorValue, ColorValue, ColorValue, ColorValue]
ColorDict = dict[str, ColorTuple]

CONFIG = config.load()

# Parameter defaults
FONT_DATA = Request(
    CONFIG["metadata.paths"]["referenceFont"],
    CONFIG["metadata.fallbacks"]["referenceFont"],
)
MARK = True
COLORS = CONFIG["color.anchors"]
CLEAR = False
VERBOSE = False

# pylint: disable=invalid-name, too-many-arguments, too-many-locals


def importAnchors(
    font: Font | Path | str,
    fontData: Request | Path | str = FONT_DATA,
    mark: bool = MARK,
    colors: ColorDict | None = COLORS,
    clear: bool = CLEAR,
    verbose: bool = VERBOSE,
) -> None:
    """Import anchors from font metadata.

    :param font: font object to which the script applies.
    :param fontData: Object call or direct path to reference font
        metadata file. Defaults to :class:`~smufolib.request.Request`
        with :obj:`~smufolib.request.Request.path`
        and :obj:`~smufolib.request.Request.fallback` set to
        :ref:`[metadata.paths]` and :ref:`[metadata.fallbacks]` respective
        `referenceFont` configurations.
    :param mark: Apply defined anchor colors. Defaults to :obj:`True`
    :param colors: dict of anchorNames mapped to UFO color values to
        apply when `mark` is :obj:`True`.
    :param clear: erase preexisting anchors on append. Defaults to
        :obj:`False`.
    :param verbose: Make output verbose. Defaults to :obj:`False`.
    :raises TypeError: If any parameter value is not the expected type
        or if `mark` is :obj:`True` while `colors` is :obj:`None`.
    :raises AttributeError: if required values for :attr:`.Smufl.name`
        are missing.

    """
    print("Starting...")

    font = _normalizeFont(font)
    metadata = _normalizeJsonDict(_normalizeRequest(fontData).json())
    colors = _normalizeColorDict(colors, mark)
    sourceAnchors = metadata["glyphsWithAnchors"]

    ticks = len(font) + len(sourceAnchors) + 1
    with tqdm(total=ticks) if not verbose else nullcontext() as progressBar:
        stdUtils.verbosePrint("\nCompiling glyph data...", verbose)
        names = {}
        for glyph in font:
            if progressBar:
                progressBar.update(1)
                time.sleep(0.0001)

            if glyph.smufl.name and glyph.smufl.name in sourceAnchors:
                names[glyph.smufl.name] = glyph.name

        if not names:
            raise ValueError(
                error.generateErrorMessage(
                    "missingValue",
                    "recommendScript",
                    objectName="Smufl.name",
                    scriptName="importID",
                )
            )

        for smuflName, anchors in sourceAnchors.items():
            if progressBar:
                progressBar.update(1)
                time.sleep(0.0001)

            try:
                glyphName = names[smuflName]
            except KeyError:
                continue
            if clear:
                font[glyphName].clearAnchors()
            stdUtils.verbosePrint(
                f"\nAppending anchors to glyph '{glyphName}':",
                verbose,
            )
            for anchorName, position in anchors.items():
                position = [font.smufl.toUnits(p) for p in position]

                anchorColor = colors.get(anchorName) if mark and colors else None
                font[glyphName].appendAnchor(anchorName, position, anchorColor)
                stdUtils.verbosePrint(f"\t{anchorName}", verbose)

        font.save()
        if progressBar:
            progressBar.update(1)
            time.sleep(0.0001)

    print("\nDone.")


def main() -> None:
    """Command line entry point."""
    args = _parseArgs()
    importAnchors(
        args.font,
        fontData=args.fontData,
        mark=args.mark,
        colors=args.colors,
        clear=args.clear,
        verbose=args.verbose,
    )


def _normalizeFont(font: Font | Path | str) -> Font:
    # Convert font path to object if necessary.
    error.validateType(font, (Font, Path, str), "font")
    if isinstance(font, Font):
        return font
    return Font(font)


def _normalizeRequest(request: Request | Path | str) -> Request:
    # Convert request path to object if necessary.
    error.validateType(request, (Request, Path, str), "request")
    if isinstance(request, Request):
        return request
    return Request(request)


def _normalizeJsonDict(jsonDict: JsonDict | None) -> JsonDict:
    # Ensure `jsonDict` is not None.
    if jsonDict is None:
        raise TypeError(error.generateTypeError(jsonDict, JsonDict, "JSON file"))
    return jsonDict


def _normalizeColorDict(colorDict: ColorDict | None, mark: bool) -> ColorDict | None:
    # Normalize dict of color values.
    if colorDict is None:
        if mark:
            raise TypeError(
                error.generateTypeError(
                    value=colorDict,
                    validTypes=ColorDict,
                    objectName="colors",
                    dependencyInfo="'mark' is True",
                )
            )
        return None

    error.validateType(colorDict, dict, "colors")
    for value in colorDict.values():
        normalizers.normalizeColor(value)

    return colorDict


def _parseArgs() -> argparse.Namespace:
    # Parse command line arguments and options.
    parser = cli.commonParser(
        "font",
        description=stdUtils.getSummary(importAnchors.__doc__),
        fontData=FONT_DATA,
        mark=MARK,
        colors=COLORS,
        clear=CLEAR,
        verbose=VERBOSE,
    )
    return parser.parse_args()


if __name__ == "__main__":
    main()



"""where SMuFL meets UFO"""

from smufolib._version import __version__  # noqa: F401
from smufolib.objects.engravingDefaults import EngravingDefaults
from smufolib.objects.engravingDefaults import ENGRAVING_DEFAULTS_KEYS
from smufolib.objects.font import Font
from smufolib.objects.glyph import Glyph
from smufolib.objects.layer import Layer
from smufolib.objects.range import Range
from smufolib.objects.smufl import Smufl
from smufolib.objects.smufl import ANCHOR_NAMES
from smufolib.objects.smufl import FONT_ATTRIBUTES
from smufolib.objects.smufl import GLYPH_ATTRIBUTES
from smufolib.request import Request

__all__ = [
    "EngravingDefaults",
    "ENGRAVING_DEFAULTS_KEYS",
    "Font",
    "Glyph",
    "Layer",
    "Range",
    "Smufl",
    "ANCHOR_NAMES",
    "FONT_ATTRIBUTES",
    "GLYPH_ATTRIBUTES",
    "Request",
]



import unittest
from smufolib.converters import (
    convertMeasurement,
    toDecimal,
    toUniHex,
    toUniName,
    toKebab,
    toNumber,
)

# pylint: disable=C0115, C0116, C0103


class Converters(unittest.TestCase):
    def test_convertMeasurement(self):
        self.assertEqual(
            convertMeasurement(125, targetUnit="spaces", unitsPerEm=1000), 0.5
        )
        self.assertEqual(
            convertMeasurement(0.5, targetUnit="units", unitsPerEm=1000), 125
        )

        with self.assertRaises(TypeError):
            convertMeasurement((125,), targetUnit="spaces", unitsPerEm=1000)
        with self.assertRaises(ValueError):
            convertMeasurement(125, targetUnit="something else", unitsPerEm=1000)

    def test_toDecimal(self):
        for value in ("U+E00C", "uE00C", "uniE00C"):
            self.assertEqual(toDecimal(value), 57356)
            with self.assertRaises(ValueError):
                toDecimal(value[1:])
        with self.assertRaises(TypeError):
            toDecimal(57356)

    def test_toUniHex(self):
        self.assertEqual(toUniHex(57344), "U+E000")
        with self.assertRaises(TypeError):
            toUniHex("57344")
        with self.assertRaises(ValueError):
            toUniHex(2000000)

    def test_toUniName(self):
        self.assertEqual(toUniName(57344), "uniE000")
        self.assertEqual(toUniName("U+E000"), "uniE000")
        self.assertEqual(toUniName(57344, short=True), "uE000")
        with self.assertRaises(TypeError):
            toUniName([57344])
        with self.assertRaises(ValueError):
            toUniName("57344")
        with self.assertRaises(ValueError):
            toUniName(2000000)

    def test_toKebab(self):
        self.assertEqual(toKebab("helloWorld!"), "hello-world!")
        # self.assertEqual(toKebab('setID'), 'set-id')
        with self.assertRaises(TypeError):
            toKebab(["485937"])

    def test_toNumber(self):
        self.assertEqual(toNumber("4598"), 4598)
        self.assertEqual(toNumber("0x4598"), 0x4598)
        self.assertEqual(toNumber("45.8"), 45.8)
        with self.assertRaises(TypeError):
            toNumber(["485937"])
        with self.assertRaises(ValueError):
            toNumber("456h")



import unittest
from smufolib import Font, Glyph
from smufolib.normalizers import (
    normalizeFont,
    normalizeClasses,
    normalizeDescription,
    normalizeDesignSize,
    normalizeSizeRange,
    normalizeSmufl,
    normalizeSmuflName,
    normalizeEngravingDefaults,
    normalizeEngravingDefaultsAttr,
    normalizeRequest,
    normalizeRequestPath,
)


class Normalizers(unittest.TestCase):
    faultyNames = ["", "smufl-name42", "SmuflName42"]
    font = Font()
    for name in ["uniE001", "uniE002", "uniE003", "uniE004"]:
        font[name] = Glyph()

    def test_normalizeFont_with_invalid_type(self):
        glyph = self.font["uniE001"]
        with self.assertRaises(TypeError):
            normalizeFont(glyph)

    def test_normalizeClasses_with_valid_input(self):
        self.assertEqual(normalizeClasses(["className"]), ("className",))

    def test_normalizeClasses_with_invalid_type(self):
        with self.assertRaises(TypeError):
            normalizeClasses("className")

    def test_normalizeClasses_with_invalid_values(self):
        for faulty_value in [
            [None],
            ["className", "className"],
            [""],
            ["smufl-name42"],
            ["SmuflName42"],
        ]:
            with self.assertRaises(ValueError):
                normalizeClasses(faulty_value)

    def test_normalizeDescription_with_invalid_type(self):
        with self.assertRaises(TypeError):
            normalizeDescription(0)

    def test_normalizeDescription_with_invalid_value(self):
        with self.assertRaises(ValueError):
            normalizeDescription("")

    def test_normalizeDesignSize_with_invalid_type(self):
        with self.assertRaises(TypeError):
            normalizeDesignSize("24")

    def test_normalizeDesignSize_with_invalid_value(self):
        with self.assertRaises(ValueError):
            normalizeDesignSize(9)

    def test_normalizeSizeRange_with_valid_input(self):
        self.assertEqual(normalizeSizeRange([10, 36]), (10, 36))

    def test_normalizeSizeRange_with_invalid_type(self):
        with self.assertRaises(TypeError):
            normalizeSizeRange(10)

    def test_normalizeSizeRange_with_invalid_values(self):
        for faulty_value in [[24], [36, 10], [9, 36]]:
            with self.assertRaises(ValueError):
                normalizeSizeRange(faulty_value)

    def test_normalizeSmufl_with_invalid_type(self):
        with self.assertRaises(TypeError):
            normalizeSmufl(self.font)

    def test_normalizeSmuflName_with_invalid_type(self):
        with self.assertRaises(TypeError):
            normalizeSmuflName(0)

    def test_normalizeSmuflName_with_invalid_values(self):
        for name in self.faultyNames:
            with self.assertRaises(ValueError):
                normalizeSmuflName(name)

    def test_normalizeEngravingDefaults_with_invalid_type(self):
        with self.assertRaises(TypeError):
            normalizeEngravingDefaults(self.font)

    def test_normalizeEngravingDefaultsAttr_with_invalid_values(self):
        for wrongAttr, wrongVal in [
            ("attribute", "23"),
            ("attribute", [23]),
            ("textFontFamily", [23]),
        ]:
            with self.assertRaises((TypeError, AttributeError, ValueError)):
                normalizeEngravingDefaultsAttr(wrongAttr, wrongVal)
        with self.assertRaises(ValueError):
            normalizeEngravingDefaultsAttr("textFontFamily", [""])

    def test_normalizeRequest_with_invalid_type(self):
        with self.assertRaises(TypeError):
            normalizeRequest("path")

    def test_normalizeRequestPath_with_invalid_type(self):
        with self.assertRaises(TypeError):
            normalizeRequestPath(["path"], parameter="path")



import unittest
from argparse import (
    ArgumentParser,
    HelpFormatter,
    RawTextHelpFormatter,
    ArgumentDefaultsHelpFormatter,
)
from smufolib.cli import commonParser, createHelpFormatter


class TestCLI(unittest.TestCase):
    def test_commonParser_basic(self):
        parser = commonParser(clear=True, addHelp=False)
        args = parser.parse_args(["--clear"])
        self.assertTrue(args.clear)

    def test_commonParser_as_parent(self):
        args = commonParser("color", clear=True, addHelp=False)
        parser = ArgumentParser(parents=[args], description="showcase commonParser")
        parser.add_argument(
            "-O",
            "--include-optionals",
            action="store_true",
            help="include optional glyphs",
            dest="includeOptionals",
        )
        parsed_args = parser.parse_args(
            ["1", "2", "3", "4", "--clear", "--include-optionals"]
        )
        self.assertEqual(parsed_args.color, [1, 2, 3, 4])
        self.assertTrue(parsed_args.clear)
        self.assertTrue(parsed_args.includeOptionals)

    def test_commonParser_with_custom_helpers(self):
        custom_helpers = {"color": "custom help for color"}
        parser = commonParser("color", clear=True, customHelpers=custom_helpers)
        args = parser.parse_args(["1", "2", "3", "4", "--clear"])
        self.assertEqual(args.color, [1, 2, 3, 4])
        self.assertTrue(args.clear)
        help_message = parser.format_help()
        self.assertIn("custom help for color", help_message)

    def test_commonParser_argument_conflict(self):
        with self.assertRaises(ValueError):
            commonParser("color", color=[1, 2, 3, 4])

    def test_createHelpFormatter_single(self):
        formatter_class = createHelpFormatter("RawTextHelpFormatter")
        self.assertTrue(issubclass(formatter_class, HelpFormatter))
        self.assertTrue(issubclass(formatter_class, RawTextHelpFormatter))

    def test_createHelpFormatter_multiple(self):
        formatter_class = createHelpFormatter(
            ("RawTextHelpFormatter", "ArgumentDefaultsHelpFormatter")
        )
        self.assertTrue(issubclass(formatter_class, HelpFormatter))
        self.assertTrue(issubclass(formatter_class, RawTextHelpFormatter))
        self.assertTrue(issubclass(formatter_class, ArgumentDefaultsHelpFormatter))

    def test_createHelpFormatter_invalid_type(self):
        with self.assertRaises(TypeError) as context:
            createHelpFormatter(123)
        self.assertIn("Expected 'formatters' to be of type", str(context.exception))

    def test_createHelpFormatter_invalid_formatter(self):
        with self.assertRaises(ValueError) as context:
            createHelpFormatter("InvalidFormatter")
        self.assertIn("InvalidFormatter", str(context.exception))



import unittest
from smufolib.objects.engravingDefaults import ENGRAVING_DEFAULTS_KEYS


class TestEngravingDefaults(unittest.TestCase):
    def setUp(self):
        # Create generic objects
        # pylint: disable=E1101
        self.font, _ = self.objectGenerator("font")
        self.smufl, _ = self.objectGenerator("smufl")
        self.engravingDefaults, _ = self.objectGenerator("engravingDefaults")

        # Assign font to engravingDefaults
        self.smufl.font = self.font
        self.font.info.familyName = "testFont"
        self.engravingDefaults.smufl = self.smufl

        # Assign glyph to font
        self.font.newGlyph("testGlyph")
        self.glyph = self.font["testGlyph"]

    def test_initialization(self):
        self.assertIsInstance(self.engravingDefaults, type(self.engravingDefaults))
        self.assertIs(self.engravingDefaults.smufl, self.smufl)

    def test_attributes(self):
        testFonts = ("font1", "font2")
        for attr in ENGRAVING_DEFAULTS_KEYS:
            with self.subTest(attr=attr):
                if attr == "textFontFamily":
                    setattr(self.engravingDefaults, attr, testFonts)
                    self.assertEqual(getattr(self.engravingDefaults, attr), testFonts)
                else:
                    setattr(self.engravingDefaults, attr, 1.0)
                    self.assertEqual(getattr(self.engravingDefaults, attr), 1.0)

    def test_set_attributes_no_font(self):
        # pylint: disable=E1101
        engravingDefaults = self.objectGenerator("engravingDefaults")
        with self.assertRaises(AttributeError):
            engravingDefaults.stemThickness = 1.0

    def test_clear(self):
        self.font.lib["com.smufolib.engravingDefaults"] = {"stemThickness": 1.0}
        self.engravingDefaults.clear()
        self.assertIsNone(self.engravingDefaults.stemThickness)

    def test_items(self):
        self.engravingDefaults.stemThickness = 1.0
        items = self.engravingDefaults.items()
        self.assertIsInstance(items, dict)
        self.assertIn("stemThickness", items)
        self.assertEqual(items["stemThickness"], 1.0)

    def test_keys(self):
        keys = self.engravingDefaults.keys()
        self.assertIsInstance(keys, list)
        self.assertIn("stemThickness", keys)

    def test_update(self):
        update_dict = {"stemThickness": 1.0, "beamThickness": 2.0}
        self.engravingDefaults.update(update_dict)
        self.assertEqual(self.engravingDefaults.stemThickness, 1.0)
        self.assertEqual(self.engravingDefaults.beamThickness, 2.0)

    def test_update_with_spaces(self):
        self.font.info.unitsPerEm = 1000
        self.engravingDefaults.spaces = True
        update_dict = {"stemThickness": 1.0, "beamThickness": 2.0}
        self.engravingDefaults.update(update_dict)
        self.assertEqual(
            self.font.lib["com.smufolib.engravingDefaults"]["stemThickness"], 250
        )
        self.assertEqual(
            self.font.lib["com.smufolib.engravingDefaults"]["beamThickness"], 500
        )

    def test_values(self):
        self.engravingDefaults.stemThickness = 1.0
        values = self.engravingDefaults.values()
        self.assertIsInstance(values, list)
        self.assertIn(1.0, values)

    def test_round(self):
        self.engravingDefaults.stemThickness = 1.5
        self.engravingDefaults.round()
        self.assertEqual(self.engravingDefaults.stemThickness, 2.0)

    def test_spaces(self):
        self.font.info.unitsPerEm = 1000
        self.assertFalse(self.engravingDefaults.spaces)
        self.smufl.spaces = True
        self.assertTrue(self.smufl.spaces)

    def test_smufl(self):
        self.assertIs(self.engravingDefaults.smufl, self.smufl)

    def test_font(self):
        self.assertIs(self.engravingDefaults.font, self.font)

    def test_glyph(self):
        self.assertIs(self.glyph.smufl.engravingDefaults.glyph, self.glyph)

    def test_layer(self):
        self.assertIs(self.glyph.smufl.engravingDefaults.layer, self.glyph.layer)



"""Utility functions for error message generation and type validation.

This module provides functions to generate and validate error messages,
check types, and suggest corrections for invalid values. It includes
a dictionary of error message templates to ensure streamlined and
consistent error reporting.

"""

from __future__ import annotations
from types import UnionType
from typing import Any, get_args
import difflib

# pylint: disable=C0103

#: Dictionary of error message templates.
ERROR_TEMPLATES: dict[str, str] = {
    "alphanumericValue": "The value for '{objectName}' must be alphanumeric.",
    "argumentConflict": "The option '{key}' is already added as positional argument or flag.",
    "attributeError": "'{objectName}' has no attribute '{attribute}'.",
    "dependentTypeError": "Expected '{objectName}' to be of type {validTypes} when {dependencyInfo}, but got {value}.",
    "dependenItemsTypeError": "Items in '{objectName}' must be {validTypes} when {dependencyInfo}, not {value}.",
    "duplicateFlags": "Arguments '{argument1}' and '{argument2}' have duplicate short flag: {flag}.",
    "duplicateItems": "Items in '{objectName}' cannot be duplicates.",
    "emptyValue": "The value for '{objectName}' cannot be empty.",
    "emptyValueItems": "Value items for '{objectName}' cannot be empty.",
    "fileNotFound": "The file or directory for '{objectName}' does not exist.",
    "invalidFormat": "The value for '{objectName}' is not correctly formatted.",
    "invalidInitialCharacter": "The value for '{objectName}' must start with a lowercase letter or number.",
    "itemsTypeError": "Items in '{objectName}' must be {validTypes}, not {value}.",
    "itemsValueError": "Invalid value for item in '{objectName}': {value}.",
    "missingExtension": "The value for '{objectName}' must have a {extension} extension.",
    "missingDependencyError": "Cannot set '{objectName}' because '{dependency}' is None.",
    "missingValue": "Required values for '{objectName}' are missing.",
    "nonIncreasingRange": "The values in '{objectName}' must form an increasing range.",
    "notImplementedError": "The '{objectName}' subclass does not implement this method.",
    "numericValue": "The value for '{objectName}' must be numeric.",
    "recommendScript": "Consider running the script '{scriptName}' before the current process.",
    "serializationError": "Error serializing JSON data or writing to the file.",
    "singleItem": "'{objectName}' must contain a value pair.",
    "suggestion": "Did you mean '{suggestion}'?",
    "typeError": "Expected '{objectName}' to be of type {validTypes}, but got {value}.",
    "unicodeOutOfRange": "The value for '{objectName}' is outside the Unicode range (U+0000 – U+10FFFF).",
    "urlError": "Could not connect to URL: {url}.",
    "valueError": "Invalid value for '{objectName}': {value}.",
    "valueTooHigh": "The value for '{objectName}' must be {value} or lower.",
    "valueTooLow": "The value for '{objectName}' must be {value} or higher.",
}


class URLWarning(Warning):
    """URL connection failure warning."""


def generateErrorMessage(*templateNames: str, **kwargs) -> str:
    """Generate an error message from a template and keyword arguments.

    :param errorTemplate: The error message template string, which
        should contain placeholders for keyword arguments.
    :param kwargs: Arbitrary keyword arguments corresponding to the
        placeholders in the template string.
    :raises KeyError: If a placeholder in the template does not have a
        corresponding keyword argument.

    Example::

        >>> generateErrorMessage('alphanumericValue', objectName='unicode')
        "The value for 'unicode' must be alphanumeric."

        >>> generateErrorMessage('typeError', objectName='index', validTypes='int', value='str')
        "Expected 'index' to be of type int, but got str."

    """
    messages = [ERROR_TEMPLATES[n].format(**kwargs) for n in templateNames]
    return " ".join(messages)


def generateTypeError(
    value: Any,
    validTypes: type | tuple[type, ...] | UnionType,
    objectName: str,
    dependencyInfo: str | None = None,
    items: bool = False,
) -> str:
    """Generate a :class:`TypeError` message.

    This function generates an error message based on the number of
    valid types. By default, the message is generated from
    :obj:`ERROR_TEMPLATES`:``'typeError'``.

    If `dependencyInfo` in not :obj:`None` (or empty) and `items`
    is :obj:`False`, :obj:`ERROR_TEMPLATES`:``'dependentTypeError'``
    is used. If `items` is :obj:`True` and `dependencyInfo` is :obj:`None`,
    :obj:`ERROR_TEMPLATES`:``'itemsTypeError'`` is used. If both
    `dependencyInfo` is not :obj:`None` and `items` is :obj:`True`,
    :obj:`ERROR_TEMPLATES`:``'dependentItemsTypeError'`` is used.

    :param value: The value to be validated.
    :param validTypes: A :class:`tuple` or :class:`list` of valid types.
    :param objectName: The name of the object being validated.
    :param dependencyInfo: Additional substring for dependent type errors.
        template. Defaults to :obj:`None`
    :param items: Whether to use items-specific template. Defaults
        to :obj:`False`.
    :raises TypeError: If `value` does not match any of the valid
        types.
    :raises ValueError: If `items` is :obj:`True` and `value` is not
        an iterable.

    Example::

        >>> generateTypeError(123, (str,), 'path')
        Expected 'path' to be of type str, but got int.

        >>> generateTypeError(123, (str, Path), 'path')
        Expected 'path' to be of type str or Path, but got int.

        >>> generateTypeError(123, (str, Path, Request), 'path')
        Expected 'path' to be of type str, Path or Request, but got int.

    """

    typeNames = _listTypes(validTypes)
    valueType = type(value).__name__

    if items:
        template = "itemsTypeError"
    elif dependencyInfo:
        template = "dependentTypeError"
    elif items and dependencyInfo:
        template = "dependenItemsTypeError"
    else:
        template = "typeError"

    return generateErrorMessage(
        template,
        validTypes=typeNames,
        objectName=objectName,
        dependencyInfo=dependencyInfo,
        value=valueType,
    )


def validateType(
    value: Any,
    validTypes: type | tuple[type, ...] | UnionType,
    objectName: str,
    items=False,
) -> None:
    """Validate if a value matches any of the valid types.

    :param value: The value to be validated.
    :param validTypes: A single valid type or :class:`tuple` of valid
        types.
    :param objectName: The name of the object being validated.
    :param items: Whether to validate `value` items rather than `value`
        itself. Defaults to :obj:`False`.
    :raises TypeError: If `value` does not match any of the valid
        types.
    :raises ValueError: If  `items` is :obj:`True` and any `value` item
        does not match any of the valid types.

    Example::

        >>> validateType(123, str, 'glyphName')
        Traceback (most recent call last):
        ...
        TypeError: Expected 'glyphName' to be of type str, but got int.

        >>> for item in ['uniE000', 1]:
        ...    validateType(item, str, 'glyphNames', items=True)
        Traceback (most recent call last):
        ...
        ValueError: Items in 'glyphNames' must be str, not int.

    """
    if isinstance(validTypes, type):
        validTypes = (validTypes,)

    elif isinstance(validTypes, UnionType):
        validTypes = get_args(validTypes)

    if not isinstance(value, validTypes):
        if items:
            raise ValueError(
                generateTypeError(value, validTypes, objectName, items=items)
            )
        else:
            raise TypeError(generateTypeError(value, validTypes, objectName))


def suggestValue(
    value: str,
    possibilities: list[str] | tuple[str, ...],
    objectName: str,
    cutoff: float = 0.6,
    items=False,
) -> str:
    """Validate value and uggests a valid close match.

    If `items` is :obj:`True`, an alternate error message template is
    used to validate the values of items within an :term:`iterable`
    value rather than the value itself.

    :param value: The value to be validated.
    :param possibilities: A list of valid possibilities for the value.
    :param objectName: The name of the object being validated.
    :param cutoff: A float value representing the cutoff threshold for
        similarity matches. Defaults to ``0.6``.
    :param items: Whether to use items-specific template. Defaults
        to :obj:`False`.
    :raises ValueError: If the provided value is not found in
        `possibilities` or if `items` is :obj:`True` and `value` is
        not an iterable.

    Example::

        >>> suggestValue('spiltStemUpSE', ['splitStemUpSE', 'splitStemUpSW'],
        ... 'anchorName', cutoff=0.5)
        "Invalid value for 'anchorName': spiltStemUpSE. Did you mean 'splitStemUpSE'?"

    """
    if value in possibilities:
        return value
    closeMatches = difflib.get_close_matches(value, possibilities, 1, cutoff)
    if closeMatches:
        suggestion = closeMatches[0]
        template = "itemsValueError" if items else "valueError"
        raise ValueError(
            generateErrorMessage(
                template,
                "suggestion",
                objectName=objectName,
                value=value,
                suggestion=suggestion,
            )
        )

    raise ValueError(
        generateErrorMessage("valueError", objectName=objectName, value=value)
    )


def _listTypes(types: type | tuple[type, ...] | UnionType) -> str:
    # Represent a series of object type names as a string.

    if isinstance(types, type):
        types = (types,)

    elif isinstance(types, UnionType):
        types = get_args(types)

    if len(types) > 1:
        typeNames = (
            ", ".join(t.__name__ for t in types[:-1]) + f" or {types[-1].__name__}"
        )
    else:
        typeNames = types[0].__name__
    return typeNames



import unittest
from smufolib import converters


class TestSmufl(unittest.TestCase):
    def setUp(self):
        # Create generic objects
        # pylint: disable=E1101
        self.font, _ = self.objectGenerator("font")
        self.smufl, _ = self.objectGenerator("smufl")
        self.engravingDefaults, _ = self.objectGenerator("engravingDefaults")
        self.glyph, _ = self.objectGenerator("glyph")
        self.range, _ = self.objectGenerator("range")

        # pylint: enable=E1101
        # Create layer and assign to font
        self.layer = self.font.newLayer("testLayer")
        self.font.defaultLayer = self.layer

        # Generate different types of glyphs to layer
        self.recommended1, self.recommended2 = self.generate_recommended_glyphs()
        self.ligature = self.generate_ligature()
        self.ligature_components = self.generate_ligature_components(self.ligature)
        self.optional = self.generate_optional()
        self.salt = self.generate_salt()
        self.set = self.generate_set()
        self.non_member1, self.non_member2 = self.generate_non_member_glyphs()

    # -------
    # Helpers
    # -------

    def generate_recommended_glyphs(self):
        glyph1 = self.layer.newGlyph("uniE000")
        glyph1.unicode = 0xE000
        glyph2 = self.layer.newGlyph("uniF3FF")
        glyph2.unicode = 0xF3FF
        return glyph1, glyph2

    def generate_ligature(self):
        glyph = self.layer.newGlyph("uniE000_uniE001_uniE002")
        glyph.unicode = 0xF400
        return glyph

    def generate_ligature_components(self, ligature):
        glyphs = []
        for i, component_name in enumerate(ligature.name.split("_")):
            glyph = self.layer.newGlyph(component_name)
            glyph.unicode = converters.toDecimal(component_name)
            glyph.lib["com.smufolib.name"] = f"smuflName{i}"
            glyphs.append(glyph)
        return tuple(glyphs)

    def generate_optional(self):
        glyph = self.layer.newGlyph("optionalGlyph")
        glyph.unicode = 0xF400
        return glyph

    def generate_salt(self):
        glyph = self.layer.newGlyph("uniE000.salt01")
        glyph.unicode = 0xF8FF
        return glyph

    def generate_set(self):
        glyph = self.layer.newGlyph("uniE000.ss01")
        glyph.unicode = 0xF400
        return glyph

    def generate_non_member_glyphs(self):
        glyph1 = self.layer.newGlyph("nonSmuflGlyph1")
        glyph1.unicode = 0xE000 - 1
        glyph2 = self.layer.newGlyph("nonSmuflGlyph2")
        glyph2.unicode = 0xF8FF + 1
        return glyph1, glyph2

    def set_dimensionsions(self, glyph, width, height):
        glyph.width = width
        glyph.height = height
        return glyph

    def draw_paths(self, glyph):
        pen = glyph.getPen()
        pen.moveTo((100, -10))
        pen.lineTo((100, 100))
        pen.lineTo((200, 100))
        pen.lineTo((200, 0))
        pen.closePath()
        pen.moveTo((110, 10))
        pen.lineTo((110, 90))
        pen.lineTo((190, 90))
        pen.lineTo((190, 10))
        pen.closePath()

    def assign_anchors(self, glyph):
        glyph.appendAnchor("stemUpNW", (1.5, 2.5))
        glyph.appendAnchor("nonSmuflAnchor", (3.5, 4.5))

    # ----
    # repr
    # ----

    def test_reprContents(self):
        self.smufl.font = self.font
        value = self.smufl._reprContents()  # pylint: disable=W0212
        self.assertIsInstance(value, list)
        for i in value:
            self.assertIsInstance(i, str)

    # -------
    # Parents
    # -------

    def test_glyph(self):
        self.smufl.glyph = self.glyph
        self.assertIsNotNone(self.smufl.glyph)
        self.assertIs(self.smufl.glyph, self.glyph)

    def test_parent_no_glyph(self):
        self.assertIsNone(self.smufl.font)
        self.assertIsNone(self.smufl.layer)
        self.assertIsNone(self.smufl.glyph)

    def test_font(self):
        self.smufl.font = self.font
        self.assertIsNotNone(self.smufl.font)
        self.assertIs(self.smufl.font, self.font)

    def test_no_font(self):
        self.smufl.glyph = self.glyph
        self.assertIsNone(self.smufl.font)

    # -------------
    # Font Metadata
    # -------------

    def test_get_designSize(self):
        self.smufl.font = self.font
        self.font.lib["com.smufolib.designSize"] = 1000
        self.assertEqual(self.smufl.designSize, 1000)

    def test_set_designSize(self):
        self.smufl.font = self.font
        self.smufl.designSize = 1000
        self.assertEqual(self.smufl.designSize, 1000)

    def test_designSize_removal(self):
        self.smufl.font = self.font
        self.font.lib["com.smufolib.designSize"] = 1000
        self.smufl.designSize = None
        self.assertIsNone(self.smufl.designSize)
        self.assertNotIn("com.smufolib.desgnSize", self.font.lib)

    def test_get_engravingDefaults(self):
        self.smufl.font = self.font
        self.assertIsInstance(
            self.smufl.engravingDefaults, type(self.engravingDefaults)
        )

    def test_set_engravingDefaults(self):
        pass

    def test_get_sizeRange(self):
        self.smufl.font = self.font
        self.font.lib["com.smufolib.sizeRange"] = (10, 36)
        self.assertEqual(self.smufl.sizeRange, (10, 36))

    def test_get_sizeRange_no_font(self):
        self.assertIsNone(self.smufl.sizeRange)

    def test_set_sizeRange(self):
        self.smufl.font = self.font
        self.smufl.sizeRange = (10, 36)
        self.assertEqual(self.smufl.sizeRange, (10, 36))

    def test_sizeRange_removal(self):
        self.smufl.font = self.font
        self.font.lib["com.smufolib.sizeRange"] = (10, 36)
        self.smufl.sizeRange = None
        self.assertEqual(self.smufl.sizeRange, None)
        self.assertNotIn("com.smufolib.sizeRange", self.font.lib)

    # --------------
    # Glyph metadata
    # --------------

    def test_alternates(self):
        self.assertEqual(self.optional.smufl.alternates, ())
        self.assertEqual(
            self.recommended1.smufl.alternates,
            (
                {"codepoint": "U+F8FF", "name": None},
                {"codepoint": "U+F400", "name": None},
            ),
        )

    def test_anchors(self):
        self.assign_anchors(self.glyph)
        self.smufl.glyph = self.glyph
        self.assertEqual(self.smufl.anchors, {"stemUpNW": (1.5, 2.5)})

    def test_bBox(self):
        self.draw_paths(self.glyph)
        self.smufl.glyph = self.glyph
        self.assertEqual(self.smufl.bBox, {"bBoxSW": (100, -10), "bBoxNE": (200, 100)})

    def test_get_codepoint(self):
        self.smufl.glyph = self.glyph
        self.glyph.unicode = 0xE111
        self.assertEqual(self.smufl.codepoint, "U+E111")

    def test_set_codepoint(self):
        self.smufl.glyph = self.glyph
        self.smufl.codepoint = "U+E111"
        self.assertEqual(self.smufl.codepoint, "U+E111")

    def test_componentGlyphs(self):
        self.assertEqual(self.ligature.smufl.componentGlyphs, self.ligature_components)

    def test_componentNames(self):
        componentNames = tuple(glyph.smufl.name for glyph in self.ligature_components)
        self.assertEqual(self.ligature.smufl.componentNames, componentNames)

    def test_range(self):
        self.assertIsInstance(self.recommended1.smufl.range, type(self.range))

    def test_get_advanceWidth(self):
        self.smufl.glyph = self.glyph
        self.glyph.width = 500
        self.assertEqual(self.smufl.advanceWidth, 500)

    def test_set_advanceWidth(self):
        self.smufl.glyph = self.glyph
        self.glyph.smufl.advanceWidth = 500
        self.assertEqual(self.smufl.advanceWidth, 500)

    # --------------
    # Identification
    # --------------

    def test_get_version(self):
        self.smufl.font = self.font
        self.font.info.versionMajor = 1
        self.font.info.versionMinor = 0
        self.assertEqual(self.smufl.version, 1.0)

    def test_set_version(self):
        self.smufl.font = self.font
        self.smufl.version = 1.0
        self.assertEqual(self.smufl.version, 1.0)

    def test_get_classes(self):
        self.smufl.glyph = self.glyph
        self.glyph.lib["com.smufolib.classes"] = ["class1", "class2"]
        self.assertEqual(self.smufl.classes, ("class1", "class2"))

    def test_get_classes_no_font(self):
        self.assertIsNone(self.smufl.classes)

    def test_set_classes(self):
        self.smufl.glyph = self.glyph
        self.smufl.classes = ["class1", "class2"]
        self.assertEqual(self.smufl.classes, ("class1", "class2"))

    def test_classes_removal(self):
        self.smufl.glyph = self.glyph
        self.glyph.lib["com.smufolib.classes"] = ["class1", "class2"]
        self.smufl.classes = None
        self.assertEqual(self.smufl.classes, ())
        self.assertNotIn("com.smufolib.classes", self.glyph.lib)

    def test_get_description(self):
        self.smufl.glyph = self.glyph
        self.glyph.lib["com.smufolib.description"] = "Test description"
        self.assertEqual(self.smufl.description, "Test description")

    def test_set_description(self):
        self.smufl.glyph = self.glyph
        self.smufl.description = "Test description"
        self.assertEqual(self.smufl.description, "Test description")

    def test_description_removal(self):
        self.smufl.glyph = self.glyph
        self.glyph.lib["com.smufolib.description"] = "Test description"
        self.smufl.description = None
        self.assertIsNone(self.smufl.designSize)
        self.assertNotIn("com.smufolib.desgnSize", self.glyph.lib)

    def test_get_name_with_font(self):
        self.smufl.font = self.font
        self.font.info.familyName = "testName"
        self.assertEqual(self.smufl.name, "testName")

    def test_set_name_with_font(self):
        self.smufl.font = self.font
        self.smufl.name = "testName"
        self.assertEqual(self.smufl.name, "testName")

    def test_get_name_with_glyph(self):
        self.recommended1.lib["com.smufolib.name"] = "testName"
        self.assertEqual(self.recommended1.smufl.name, "testName")

    def test_set_name_with_glyph(self):
        glyph = self.recommended1
        glyph.smufl.name = "testName"
        self.assertEqual(glyph.smufl.name, "testName")
        # Test presence of com.smufolib.names
        self.assertEqual(glyph.font.lib["com.smufolib.names"], {"testName": "uniE000"})

    def test_name_with_glyph_removal(self):
        self.recommended1.lib["com.smufolib.name"] = "testName"
        self.recommended1.smufl.name = None
        self.assertIsNone(self.recommended1.smufl.name)
        self.assertNotIn("com.smufolib.name", self.recommended1.lib)
        # Test removal of com.smufolib.names
        self.assertNotIn("com.smufolib.names", self.recommended1.font.lib)

    # ----------
    # Validation
    # ----------

    def test_isLigature(self):
        self.assertTrue(self.ligature.smufl.isLigature)
        self.assertFalse(self.recommended1.smufl.isLigature)

    def test_isMember(self):
        self.assertTrue(self.recommended1.smufl.isMember)
        self.assertTrue(self.recommended2.smufl.isMember)
        self.assertFalse(self.non_member1.smufl.isMember)
        self.assertFalse(self.non_member2.smufl.isMember)

    def test_isOptional(self):
        self.assertTrue(self.optional.smufl.isOptional)
        self.assertFalse(self.recommended1.smufl.isOptional)

    def test_isRecommended(self):
        self.assertTrue(self.recommended1.smufl.isRecommended)
        self.assertTrue(self.recommended2.smufl.isRecommended)
        self.assertFalse(self.optional.smufl.isRecommended)

    def test_isSalt(self):
        self.assertTrue(self.salt.smufl.isSalt)
        self.assertFalse(self.set.smufl.isSalt)

    def test_isSet(self):
        self.assertTrue(self.set.smufl.isSet)
        self.assertFalse(self.salt.smufl.isSet)

    # -----------------------------
    # Normalization and Measurement
    # -----------------------------

    def test_round(self):
        self.set_dimensionsions(self.recommended1, 100.5, 100.5)
        self.assign_anchors(self.recommended1)
        self.recommended1.font.info.unitsPerEm = 1000

        self.recommended1.font.lib["com.smufolib.spaces"] = True
        self.recommended1.smufl.round()
        self.assertEqual(self.recommended1.smufl.advanceWidth, 100.5 / 250)
        self.assertEqual(
            self.recommended1.smufl.anchors, {"stemUpNW": (1.5 / 250, 2.5 / 250)}
        )

        self.recommended1.font.lib.pop("com.smufolib.spaces")
        self.recommended1.smufl.round()
        self.assertEqual(self.recommended1.smufl.advanceWidth, 101)
        self.assertEqual(self.recommended1.smufl.anchors, {"stemUpNW": (2, 3)})

    def test_toSpaces(self):
        self.recommended1.font.info.unitsPerEm = 1000
        self.assertEqual(self.recommended1.smufl.toSpaces(125), 0.5)

    def test_toUnits(self):
        self.recommended1.font.info.unitsPerEm = 1000
        self.assertEqual(self.recommended1.smufl.toUnits(0.499), 125)
        self.assertEqual(self.recommended1.smufl.toUnits(0.499, rounded=False), 124.75)

    def test_spaces(self):
        self.recommended1.font.info.unitsPerEm = 1000
        self.assertFalse(self.recommended1.smufl.spaces)
        self.recommended1.smufl.spaces = True
        self.assertTrue(self.recommended1.smufl.spaces)
        with self.assertRaises(AttributeError):
            self.smufl.spaces = True
            self.recommended1.font.info.unitsPerEm = None

    # -----
    # Other
    # -----

    def test_base(self):
        self.assertEqual(self.salt.smufl.base.name, self.recommended1.name)
        self.assertEqual(self.recommended1.smufl.base.name, self.recommended1.name)

    def test_base_no_font(self):
        self.smufl.glyph = self.glyph
        self.glyph.name = "uniE000.salt"
        self.assertIsNone(self.glyph.smufl.base)

    def test_base_no_glyph_name(self):
        pass

    def test_base_not_in_font(self):
        glyph = self.recommended1
        glyph.font.removeGlyph(self.recommended1.name)
        self.assertIsNone(glyph.smufl.base)

    def test_findGlyph(self):
        self.recommended1.smufl.name = "smuflName1"
        self.assertIsInstance(self.font.smufl.findGlyph("smuflName1"), type(self.glyph))
        self.assertIsNone(self.font.smufl.findGlyph("smuflName2"))

    def test_suffix(self):
        self.assertEqual(self.salt.smufl.suffix, "salt01")



import unittest
from pathlib import Path
from smufolib import Request
from smufolib.error import (
    generateErrorMessage,
    generateTypeError,
    validateType,
    suggestValue,
    _listTypes,
)


class TestError(unittest.TestCase):
    def test_generateErrorMessage(self):
        message = generateErrorMessage("alphanumericValue", objectName="unicode")
        expected_message = "The value for 'unicode' must be alphanumeric."
        self.assertEqual(message, expected_message)

        message = generateErrorMessage(
            "typeError", objectName="index", validTypes="int", value="str"
        )
        expected_message = "Expected 'index' to be of type int, but got str."
        self.assertEqual(message, expected_message)

    def test_generateTypeError(self):
        message = generateTypeError(123, (str,), "path")
        expected_message = "Expected 'path' to be of type str, but got int."
        self.assertEqual(message, expected_message)

        message = generateTypeError(123, (str, Path), "path")
        expected_message = "Expected 'path' to be of type str or Path, but got int."
        self.assertEqual(message, expected_message)

        message = generateTypeError(123, (str, Path, Request), "path")
        expected_message = (
            "Expected 'path' to be of type str, Path or Request, but got int."
        )
        self.assertEqual(message, expected_message)

    def test_validateType(self):
        with self.assertRaises(TypeError) as context:
            validateType(123, str, "glyphName")
        self.assertEqual(
            str(context.exception),
            "Expected 'glyphName' to be of type str, but got int.",
        )

        validateType("uniE001", str, "glyphNames", items=True)

        with self.assertRaises(ValueError) as context:
            validateType(["uniE000", 1], str, "glyphNames", items=True)
        self.assertEqual(
            str(context.exception), "Items in 'glyphNames' must be str, not list."
        )

    def test_suggestValue(self):
        with self.assertRaises(ValueError) as context:
            suggestValue(
                "spiltStepUpSE",
                ["splitStemUpSE", "splitStemUpSW"],
                "anchorName",
                cutoff=0.5,
            )
        self.assertEqual(
            str(context.exception),
            "Invalid value for 'anchorName': spiltStepUpSE. Did you mean 'splitStemUpSE'?",
        )

        with self.assertRaises(ValueError) as context:
            suggestValue("invalidValue", ["validValue1", "validValue2"], "objectName")
        self.assertEqual(
            str(context.exception),
            "Invalid value for 'objectName': invalidValue. Did you mean 'validValue2'?",
        )

    def test_listTypes(self):
        self.assertEqual(_listTypes(str), "str")
        self.assertEqual(_listTypes((str, int)), "str or int")
        self.assertEqual(_listTypes((str, int, float)), "str, int or float")



import unittest
from unittest.mock import patch


class TestRange(unittest.TestCase):
    def setUp(self):
        # Using objectGenerator to create actual objects
        # pylint: disable=E1101
        self.font, _ = self.objectGenerator("font")
        self.smufl, _ = self.objectGenerator("smufl")
        self.layer, _ = self.objectGenerator("layer")
        self.range, _ = self.objectGenerator("range")

        # pylint: enable=E1101
        # Create layer and assign to font
        self.layer = self.font.newLayer("testLayer")
        self.font.defaultLayer = self.layer
        self.range.smufl = self.smufl
        self.glyph1 = self.generateGlyph("uniE080", 0xE080, "timeSig0")
        self.glyph2 = self.generateGlyph("uniE081", 0xE081, "timeSig1")
        self.smufl.glyph = self.glyph1

    def generateGlyph(self, name, unicode, smufl_name):
        glyph = self.font.newGlyph(name)
        glyph.unicode = unicode
        glyph.smufl.name = smufl_name
        return glyph

    def test_smufl(self):
        self.assertEqual(self.range.smufl, self.smufl)

    def test_glyph(self):
        self.assertEqual(self.range.glyph, self.glyph1)

    def test_font(self):
        self.assertEqual(self.range.font, self.font)

    def test_layer(self):
        self.assertEqual(self.range.layer, self.layer)

    @patch(
        "smufolib.objects.range.METADATA",
        {
            "timeSignatures": {
                "description": "Time signatures",
                "glyphs": ["timeSig0", "timeSig1"],
                "range_start": "U+E080",
                "range_end": "U+E09F",
            }
        },
    )
    def test_name(self):
        self.assertEqual(self.range.name, "timeSignatures")

    def test_description(self):
        self.assertEqual(self.range.description, "Time signatures")

    def test_start(self):
        self.assertEqual(self.range.start, "U+E080")

    def test_end(self):
        self.assertEqual(self.range.end, "U+E09F")

    def test_glyphs(self):
        self.assertEqual(self.range.glyphs, (self.glyph1, self.glyph2))



import unittest
from unittest.mock import patch, mock_open
from configparser import ConfigParser
import errno
from smufolib.config import load, _readConfigFile, _selectPath, _parse


class TestConfig(unittest.TestCase):
    # pylint: disable=W0613

    @patch("smufolib.config._readConfigFile")
    def test_load(self, mock_readConfigFile):
        mock_config = ConfigParser()
        mock_config.add_section("request")
        mock_config.set("request", "encoding", "utf-8")
        mock_config.set("request", "warn", "True")
        mock_readConfigFile.return_value = mock_config

        result = load()
        self.assertEqual(result["request"]["encoding"], "utf-8")
        self.assertEqual(result["request"]["warn"], True)

    @patch(
        "builtins.open", new_callable=mock_open, read_data="[section]\noption=value\n"
    )
    @patch("smufolib.config._selectPath", return_value="dummy_path")
    def test_readConfigFile(self, mock_selectPath, mock_open_file):
        result = _readConfigFile("dummy_path")
        self.assertIsInstance(result, ConfigParser)
        self.assertTrue(result.has_section("section"))
        self.assertEqual(result.get("section", "option"), "value")

    @patch("os.getenv", return_value=None)
    @patch(
        "pathlib.Path.exists", side_effect=[False, False, True]
    )  # Only three checks needed
    def test_selectPath(self, mock_exists, mock_getenv):
        result = _selectPath(None)
        self.assertTrue(result.endswith("smufolib.cfg"))

    @patch("os.getenv", return_value=None)
    @patch("pathlib.Path.exists", return_value=False)
    def test_selectPath_not_found(self, mock_exists, mock_getenv):
        with self.assertRaises(FileNotFoundError) as context:
            _selectPath(None)
        self.assertEqual(context.exception.errno, errno.ENOENT)

    def test_parse(self):
        config = ConfigParser()
        config.add_section("section")
        config.set("section", "int_option", "1")
        config.set("section", "float_option", "1.1")
        config.set("section", "bool_option", "True")
        config.set("section", "str_option", "string")
        config.set("section", "tuple_option", "(1, 2.2, string)")

        self.assertEqual(_parse(config, "section", "int_option"), 1)
        self.assertEqual(_parse(config, "section", "float_option"), 1.1)
        self.assertEqual(_parse(config, "section", "bool_option"), True)
        self.assertEqual(_parse(config, "section", "str_option"), "string")
        self.assertEqual(_parse(config, "section", "tuple_option"), (1, 2.2, "string"))
